

import typeguard

from cloud_governance.cloud_resource_orchestration.clouds.common.abstract_tagging_operations import \
    AbstractTaggingOperations
from cloud_governance.common.clouds.azure.compute.resource_group_operations import ResourceGroupOperations
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


class AzureTaggingOperations(AbstractTaggingOperations):
    """
    This class performs the tagging operations on AWS
    """

    def __init__(self):
        super(AbstractTaggingOperations).__init__()
        self.__resource_group_operations = ResourceGroupOperations()

    @logger_time_stamp
    def tag_resources_list(self, resources_list: list, update_tags_dict: dict):
        """
        This method updates the tags to the resources
        :param resources_list:
        :param update_tags_dict:
        :return:
        """
        pass

    @typeguard.typechecked
    @logger_time_stamp
    def get_resources_list(self, tag_name: str, tag_value: str = ''):
        """
        This method returns all the resources having the tag_name and tag_value
        :param tag_name:
        :param tag_value:
        :return:
        """
        filter_values = f"``$filter=tagName eq '{tag_name}' and tagValue eq '{tag_value}'``"
        return self.__resource_group_operations.get_all_resources(filter=filter_values)




from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class S3Inactive(AWSPolicyOperations):
    """
    This class sends an alert mail for empty bucket to the user after 4 days and delete after 7 days.
    """

    RESOURCE_ACTION = 'Delete'

    def __init__(self):
        super().__init__()
        self.__global_active_cluster_ids = self._get_global_active_cluster_ids()

    def run_policy_operations(self):
        """
        This method returns all Empty buckets
        :return:
        :rtype:
        """
        empty_buckets = []
        s3_buckets = self._s3operations.list_buckets()
        for bucket in s3_buckets:
            bucket_name = bucket.get('Name')
            tags = self._s3operations.get_bucket_tagging(bucket_name)
            cleanup_result = False
            cluster_tag = self._get_cluster_tag(tags=tags)
            cleanup_days = 0
            s3_contents = self._s3operations.get_bucket_contents(bucket_name=bucket_name)
            if cluster_tag not in self.__global_active_cluster_ids and len(s3_contents) == 0:
                cleanup_days = self.get_clean_up_days_count(tags=tags)
                cleanup_result = self.verify_and_delete_resource(resource_id=bucket_name, tags=tags,
                                                                 clean_up_days=cleanup_days)
                region = self._s3operations.get_bucket_location(bucket_name=bucket_name)
                resource_data = self._get_es_schema(resource_id=bucket_name,
                                                    user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                    skip_policy=self.get_skip_policy_value(tags=tags),
                                                    cleanup_days=cleanup_days,
                                                    dry_run=self._dry_run,
                                                    name=bucket_name,
                                                    region=region,
                                                    cleanup_result=str(cleanup_result),
                                                    resource_action=self.RESOURCE_ACTION,
                                                    cloud_name=self._cloud_name,
                                                    resource_type='EmptyBucket',
                                                    resource_state="Empty",
                                                    unit_price=0)
                empty_buckets.append(resource_data)
            if not cleanup_result:
                self.update_resource_day_count_tag(resource_id=bucket_name, cleanup_days=cleanup_days, tags=tags)

        return empty_buckets



import json
import datetime


class JsonDateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            # Serialize datetime objects to ISO 8601 format
            return obj.isoformat()
        return super(JsonDateTimeEncoder, self).default(obj)



import datetime
import os

import SoftLayer
import pandas as pd
from ibm_platform_services import UsageReportsV4
from retry import retry
from typeguard import typechecked

from cloud_governance.common.google_drive.google_drive_operations import GoogleDriveOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class IBMAccount:
    """
    This class contains IBM softlayer client and methods
    For Usage reports need to export
    export USAGE_REPORTS_URL=<SERVICE_URL>
    export USAGE_REPORTS_AUTHTYPE=iam
    export USAGE_REPORTS_APIKEY=<API_KEY>
    """

    START_DATE = 1
    END_DATE = 16
    RETRIES = 3
    DELAY = 30

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__API_USERNAME = self.__environment_variables_dict.get('IBM_API_USERNAME', '')
        self.__API_KEY = self.__environment_variables_dict.get('IBM_API_KEY', '')
        try:
            if self.__API_KEY and self.__API_USERNAME:
                self.__sl_client = SoftLayer.create_client_from_env(username=self.__API_USERNAME, api_key=self.__API_KEY, timeout=self.DELAY)
                self.short_account_id = str(self.__API_USERNAME.split('_')[0])
            if self.__environment_variables_dict.get('USAGE_REPORTS_APIKEY'):
                self.__service_client = UsageReportsV4.new_instance()
        except Exception as err:
            raise err
        self.account = self.__environment_variables_dict.get('account', '')
        self.__account_id = self.__environment_variables_dict.get('IBM_ACCOUNT_ID', '')

        self.__gsheet_id = self.__environment_variables_dict.get('SPREADSHEET_ID', '')
        self.__gsheet_client = GoogleDriveOperations()
        self.__ldap_host_name = self.__environment_variables_dict.get('LDAP_HOST_NAME', '')
        self.__ldap = LdapSearch(ldap_host_name=self.__ldap_host_name)

    def get_sl_client(self):
        """
        This method return the softlayer client
        @return:
        """
        return self.__sl_client

    @typechecked
    def __organise_user_tags(self, tags: dict):
        """
        This method organise the tags from the gsheet
        @param tags:
        @return:
        """
        user_tags = []
        for tag, value in tags.items():
            if value.strip():
                value = value.replace('/', '-')
                user_tags.append(f'{tag.strip().lower()}:{value.strip().lower()}')
        return user_tags

    @typechecked
    def get_user_tags_from_gsheet(self, username: str, user_email: str = '', file_path: str = '/tmp/'):
        """
        This method return the user tags from the gsheet
        @param user_email:
        @param username:
        @param file_path:
        @return:
        """
        file_name = os.path.join(file_path, f'{self.account}.csv')
        if not os.path.exists(file_name):
            self.__gsheet_client.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=self.account,
                                                      file_path=file_path)
        df = pd.read_csv(file_name)
        df.fillna('', inplace=True)
        if user_email:
            df.set_index('_Email', inplace=True)
            user = username.split('@')[0]
        else:
            df.set_index('User', inplace=True)
            user = username.split('_')[1].split('@')[0]
        tags = dict(df.loc[username])
        tags['User'] = user
        for key in list(tags.keys()):
            if key.startswith('_'):
                tags.pop(key)
        ldap_data = self.__ldap.get_user_details(user_name=tags['User'])
        if ldap_data:
            tags['Owner'] = ldap_data['FullName']
            tags['Manager'] = ldap_data['managerName']
        return self.__organise_user_tags(tags)

    @retry(exceptions=Exception, tries=RETRIES, delay=DELAY)
    @typechecked
    @logger_time_stamp
    def get_monthly_invoices(self, month: int, year: int):
        _filter = {
            'invoices': {
                'closedDate': {
                    'operation': 'betweenDate',
                    'options': [
                        {'name': 'startDate', 'value': [f'{month} / {self.START_DATE} / {year}']},
                        {'name': 'endDate', 'value': [f'{month} / {self.END_DATE} / {year}']}
                    ]
                }
            }
        }
        invoice_mask = "mask[id, closedDate, typeCode, createDate]"
        invoice_list = self.__sl_client.call('SoftLayer_Account', 'getInvoices', mask=invoice_mask, filter=_filter, iter=True)
        invoice_data = {}
        for invoice in invoice_list:
            if invoice.get('typeCode') == 'RECURRING':
                invoice_item_mask = f"""mask[id, createDate, recurringFee, parentId, categoryCode, description, hostName, domainName, invoiceId, resourceTableId, productItemId]"""
                invoice_items = self.__sl_client.call('SoftLayer_Billing_Invoice', 'getItems', id=invoice.get('id'), iter=True, mask=invoice_item_mask)
                invoice_data[invoice.get('id')] = invoice_items
        return invoice_data

    @retry(exceptions=Exception, tries=RETRIES, delay=DELAY)
    @logger_time_stamp
    def get_users(self):
        """
        This method returns all the users of the ibm cloud
        @return:
        """
        user_mask = "email"
        users = self.__sl_client.call('SoftLayer_Account', 'getUsers', mask=user_mask, iter=True)
        if users:
            return [user['email'].split('@')[0] for user in users]
        else:
            return []

    @typechecked
    @logger_time_stamp
    def get_invoice_data(self, month: int, year: int):
        """
        This method return invoice data
        @return:
        """
        users = self.get_users()
        data = self.get_monthly_invoices(month, year)
        hostname_data = {}
        parent_id_data = {}
        description_id_data = {}
        for invoice_id, invoices in data.items():
            for invoice in invoices:
                parent_id = invoice.get('parentId')
                item_id = invoice.get('id')
                recurring_fee = float(invoice.get('recurringFee', 0))
                description = invoice.get('description')
                username = [user for user in users if user in description]
                if not parent_id:
                    if 'hostName' in invoice:
                        hostname_data[item_id] = f"""{invoice.get('hostName')}.{invoice.get('domainName')}"""
                    else:
                        hostname_data[item_id] = f"""{invoice.get('categoryCode')}"""
                if parent_id:
                    parent_id_data[parent_id] = parent_id_data.get(parent_id, 0) + recurring_fee
                else:
                    parent_id_data[item_id] = parent_id_data.get(item_id, 0) + recurring_fee
                if username:
                    description_id_data.setdefault(parent_id, set()).add(f'{username[0]}-{"-".join(description.split()[:2])}')
        for parent_id, username in description_id_data.items():
            hostname_data[parent_id] = list(username)[0]
        combine_invoice_data = {}
        co = 0
        for key, value in parent_id_data.items():
            if value > 0:
                if hostname_data.get(key).lower() in combine_invoice_data:
                    combine_invoice_data[hostname_data.get(key).lower()]['cost'] += value
                else:
                    combine_invoice_data[hostname_data.get(key).lower()] = {
                        'fqdn': hostname_data.get(key).lower(),
                        'cost': value
                    }
                co += value
        return combine_invoice_data, users

    @logger_time_stamp
    def get_next_recurring_invoice(self):
        """This method returns the next recurring invoice"""
        invoice_items = self.__sl_client.call('SoftLayer_Account', 'getNextInvoiceTotalRecurringAmount', iter=True)
        return float(invoice_items[0]) if invoice_items else 0

    @logger_time_stamp
    def get_daily_usage(self, month: int, year: int):
        """
        This method get IBM monthly usage
        @param month:
        @param year:
        @return:
        """
        billing_month = str(year) + '-' + str(month)  # yyyy-mm
        account_summary = self.__service_client.get_account_summary(account_id=self.__account_id, billingmonth=billing_month, timeout=self.DELAY).get_result()
        return account_summary



import importlib
import inspect
import os.path
from abc import abstractmethod, ABC
from typing import Union, Callable

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.main.environment_variables import environment_variables
from cloud_governance.policy.policy_runners.elasticsearch.upload_elastic_search import UploadElasticSearch


class AbstractPolicyRunner(ABC):

    def __init__(self):
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._policy = self._environment_variables_dict.get('policy', '')
        self._account = self._environment_variables_dict.get('account', '')
        self._dry_run = self._environment_variables_dict.get('dry_run', 'yes')
        self._region = self._environment_variables_dict.get('AWS_DEFAULT_REGION', 'us-east-2')
        self._run_active_regions = self._environment_variables_dict.get('RUN_ACTIVE_REGIONS')
        self._upload_elastic_search = UploadElasticSearch()
        self._save_to_file_path = self._environment_variables_dict.get('SAVE_TO_FILE_PATH')
        self._public_cloud_name = self._environment_variables_dict.get('PUBLIC_CLOUD_NAME', '')

    def write_to_file(self, data: Union[list, dict, str]):
        """
        This method writes the data to file_path passed by the env SAVE_TO_FILE_PATH
        :param data:
        :type data:
        :return:
        :rtype:
        """
        if self._save_to_file_path:
            if os.path.exists(self._save_to_file_path):
                if data:
                    header_added = False
                    file_name = f'{self._save_to_file_path}/{self._policy}.csv'
                    with open(file_name, 'w') as file:
                        if isinstance(data, list):
                            for item in data:
                                if isinstance(item, dict):
                                    if not header_added:
                                        keys = [str(val) for val in list(item.keys())] + ["\n"]
                                        file.write(', '.join(keys))
                                        header_added = True
                                    values = [str(val) for val in list(item.values())] + ["\n"]
                                    file.write(', '.join(values))
                                else:
                                    file.write(f'{item}\n')
                        else:
                            if isinstance(data, dict):
                                if not header_added:
                                    keys = [str(val) for val in list(data.keys())] + ["\n"]
                                    file.write(', '.join(keys))
                                values = [str(val) for val in list(data.values())] + ["\n"]
                                file.write(', '.join(values))
                            else:
                                file.write(data)
                                file.write('\n')
                    logger.info(f"Written the data into the file_name: {file_name}")
            else:
                raise FileExistsError(f"FilePath not exists {self._save_to_file_path}")

    @abstractmethod
    def execute_policy(self, policy_class_name: str, run_policy: Callable, upload: bool):
        """
        This method execute the policy
        :return:
        :rtype:
        """
        raise NotImplementedError("This method is not yet implemented")

    def run(self, source: str = "", upload: bool = True):
        """
        This method starts the method operations
        :param source:
        :type source:
        :param upload:
        :type upload:
        :return:
        :rtype:
        """
        source_policy = f"{source}.{self._policy}" if source else self._policy
        logger.info(f'CloudName={self._public_cloud_name}, account={self._account}, policy={self._policy}, dry_run={self._dry_run}')
        policies_path = f'cloud_governance.policy.{self._public_cloud_name.lower()}.{source_policy}'
        cloud_policies = importlib.import_module(policies_path)
        policy_result = []

        for cls in inspect.getmembers(cloud_policies, inspect.isclass):
            if self._policy.replace('_', '').replace('-', '') == cls[0].lower():
                response = self.execute_policy(policy_class_name=cls[0], run_policy=cls[1], upload=upload)
                policy_result.extend(response)
        if self._save_to_file_path:
            self.write_to_file(data=policy_result)
        return policy_result




AWS: str = "AWS"
GCP: str = "GCP"
AZURE: str = "AZURE"
IBM: str = "IBM"
JIRA_ISSUE_NEW_STATE: str = 'New'
JIRA_ISSUE_REFINEMENT_STATE: str = 'Refinement'
JIRA_ISSUE_CLOSED_STATE: str = 'Closed'
JIRA_ISSUE_IN_PROGRESS_STATE: str = 'In Progress'
FIRST_CRO_ALERT: int = 5
SECOND_CRO_ALERT: int = 3
CLOSE_JIRA_TICKET: int = 0
DEFAULT_ROUND_DIGITS: int = 3
DATE_FORMAT: str = '%Y-%m-%d'
DURATION = 'Duration'
TICKET_ID = 'TicketId'
CLOUD_GOVERNANCE_ES_MAIL_INDEX = 'cloud-governance-mail-messages'
CRO_OVER_USAGE_ALERT = 'cro-over-usage-alert'
TIMESTAMP_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
OVER_USAGE_THRESHOLD = 0.05
SEND_ALERT_DAY = 3



from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class ZombieSnapshots(AWSPolicyOperations):
    """
    This class sends an alert mail for zombie snapshots ( AMI abandoned ) to the user after 4 days and delete after 7 days.
    """

    RESOURCE_ACTION = 'Delete'

    def __init__(self):
        super().__init__()
        self.__image_ids = self._get_ami_ids()

    def __snapshot_id_in_images(self, resource_id: str):
        """
        This method checks if snapshot_id exists in images list
        :param resource_id:
        :return:
        """
        image_snapshot_filter = {
            'Name': 'block-device-mapping.snapshot-id',
            'Values': [
                resource_id,
            ]
        }
        images = self._get_ami_ids(Filters=[image_snapshot_filter])
        return len(images) >= 1

    def run(self):
        """
        This method returns all the zombie snapshots and delete after x days
        @return:
        """
        snapshots = self._ec2_operations.get_snapshots()
        zombie_snapshots = []
        for snapshot in snapshots:
            tags = snapshot.get('Tags', [])
            resource_id = snapshot.get('SnapshotId')
            cleanup_result = False
            cluster_tag = self._get_cluster_tag(tags=tags)
            cleanup_days = 0
            if not cluster_tag and not self.__snapshot_id_in_images(resource_id):
                cleanup_days = self.get_clean_up_days_count(tags=tags)
                cleanup_result = self.verify_and_delete_resource(resource_id=resource_id, tags=tags,
                                                                 clean_up_days=cleanup_days)
                unit_price = 0
                resource_data = self._get_es_schema(resource_id=resource_id,
                                                    user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                    skip_policy=self.get_skip_policy_value(tags=tags),
                                                    cleanup_days=cleanup_days, dry_run=self._dry_run,
                                                    name=self.get_tag_name_from_tags(tags=tags, tag_name='Name'),
                                                    region=self._region,
                                                    cleanup_result=str(cleanup_result),
                                                    resource_action=self.RESOURCE_ACTION,
                                                    cloud_name=self._cloud_name,
                                                    resource_type='Snapshot',
                                                    volume_size=f"{snapshot.get('VolumeSize')} GB",
                                                    unit_price=unit_price, resource_state='Backup' if not cleanup_result else "Deleted"
                                                    )
                zombie_snapshots.append(resource_data)
            if not cleanup_result:
                self.update_resource_day_count_tag(resource_id=resource_id, cleanup_days=cleanup_days, tags=tags)
        return zombie_snapshots



import logging
from abc import ABC
from datetime import datetime, timedelta

import typeguard

from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.cost_over_usage import CostOverUsage
from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.jira.jira_operations import JiraOperations
from cloud_governance.common.logger.init_logger import handler
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class AbstractCollectCROReports(ABC):
    """
    This method collects the user/instance-id data from the cost-explorer
    """

    DEFAULT_ROUND_DIGITS = 3
    ZERO = 0
    TICKET_ID_KEY = 'ticket_id'
    TICKET_ID_VALUE = 'TicketId'
    AND = 'And'
    ALLOCATED_BUDGET = 'AllocatedBudget'

    def __init__(self):
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._account_name = self._environment_variables_dict.get('account', '')
        self._jira_operations = JiraOperations()
        self._public_cloud_name = self._environment_variables_dict.get('PUBLIC_CLOUD_NAME', '')
        self._es_index_cro = self._environment_variables_dict.get('CRO_ES_INDEX', '')
        self._ce_payer_index = self._environment_variables_dict.get('CE_PAYER_INDEX')
        self._es_operations = ElasticSearchOperations()

    def _get_account_budget_from_payer_ce_report(self):
        """
        This method returns the total account budget
        :return:
        :rtype:
        """
        raise NotImplementedError("Account Budget Not Implemented Error")

    @typeguard.typechecked
    @logger_time_stamp
    def _prepare_instance_data(self, instance_data: list, user: str, ticket_id: str, user_cost: float,
                              cost_estimation: float, ticket_opened_date: datetime):
        """
        This method returns es data to upload
        :param instance_data:
        :param user:
        :param ticket_id:
        :param user_cost:
        :param cost_estimation:
        :param ticket_opened_date:
        :return: dict data
        """
        instance_meta_data = []
        cluster_names = []
        for data in instance_data:
            if type(data.get('instance_data')) is not list:
                instance_meta_data.append(data.get('instance_data'))
            else:
                instance_meta_data.extend(data.get('instance_data'))
            if data.get('cluster_name'):
                cluster_names.append(data.get('cluster_name'))
        return {
            'cloud_name': self._public_cloud_name.upper(),
            'account_name': self._account_name,
            'region_name': instance_data[self.ZERO].get('region_name'),
            'user': user,
            'user_cro': instance_data[self.ZERO].get('user_cro'),
            'actual_cost': user_cost,
            'ticket_id': ticket_id,
            'ticket_id_state': 'in-progress',
            'estimated_cost': cost_estimation,
            'ticket_opened_date': ticket_opened_date.date(),
            'duration': int(instance_data[self.ZERO].get('duration')),
            'approved_manager': instance_data[self.ZERO].get('approved_manager'),
            'user_manager': instance_data[self.ZERO].get('manager'),
            'project': instance_data[self.ZERO].get('project'),
            'owner': instance_data[self.ZERO].get('owner'),
            self.ALLOCATED_BUDGET: self._get_account_budget_from_payer_ce_report(),
            'instance_data': instance_meta_data,
            'cluster_names': cluster_names,
        }

    @typeguard.typechecked
    @logger_time_stamp
    def _prepare_update_es_data(self, source: dict, instance_data: list, user_cost: float, cost_estimation: float):
        """
        This method updates the values of jira id data
        :param source:
        :param instance_data:
        :param user_cost:
        :param cost_estimation:
        :return: dict data
        """
        es_instance_data = source.get('instance_data', [])
        es_cluster_names = source.get('cluster_names', [])
        for instance in instance_data:
            instance_meta_data = instance.get('instance_data')
            if instance.get('cluster_name') and instance.get('cluster_name') not in es_cluster_names:
                source.setdefault('cluster_names', []).append(instance.get('cluster_name'))
            if type(instance_meta_data) is not list:
                instance_meta_data = [instance.get('instance_data')]
            for data in instance_meta_data:
                if data not in es_instance_data:
                    source.setdefault('instance_data', []).append(data)
        source['cluster_names'] = list(set(source.get('cluster_names', [])))
        source['duration'] = int(instance_data[self.ZERO].get('duration'))
        source['estimated_cost'] = round(cost_estimation, self.DEFAULT_ROUND_DIGITS)
        source['actual_cost'] = user_cost
        if instance_data[self.ZERO].get('user_cro') and source.get('user_cro') != instance_data[self.ZERO].get('user_cro'):
            source['user_cro'] = instance_data[self.ZERO].get('user_cro')
        if instance_data[self.ZERO].get('user') and source.get('user') != instance_data[self.ZERO].get('user'):
            source['user'] = instance_data[self.ZERO].get('user')
        source['timestamp'] = datetime.utcnow()
        if source.get('ticket_id_state') != 'in-progress':
            source['ticket_id_state'] = 'in-progress'
            source['approved_manager'] = instance_data[self.ZERO].get('approved_manager')
            source['user_manager'] = instance_data[self.ZERO].get('manager'),
            source['user_manager'] = instance_data[self.ZERO].get('manager'),
            source[self.ALLOCATED_BUDGET] = self._get_account_budget_from_payer_ce_report()
        return source

    def _check_value_in_es(self, tag_key: str, tag_value: str, ticket_id: str):
        """
        This method returns the bool on comparing the es tag_key value and tag_value
        :param tag_key:
        :param tag_value:
        :param ticket_id:
        :return:
        """
        es_data = self._es_operations.get_es_data_by_id(index=self._es_index_cro, id=ticket_id)
        es_tag_value = es_data.get('_source', {}).get(tag_key.lower(), '').lower()
        return es_tag_value == tag_value

    def _upload_cro_report_to_es(self, monitor_data: dict):
        """
        This method uploads the data to elastic search index and return the data
        :param monitor_data:
        :return:
        """
        raise NotImplementedError("Not Implemented")

    @typeguard.typechecked
    @logger_time_stamp
    def run(self, monitor_data: dict):
        """
        This method runs data collection methods
        :param monitor_data:
        :return:
        """
        result = self._upload_cro_report_to_es(monitor_data=monitor_data)
        return result





import json
import tempfile
from abc import ABC
from datetime import datetime

import typeguard

from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.aws_tagging_operations import AWSTaggingOperations
from cloud_governance.cloud_resource_orchestration.common.abstract_monitor_tickets import AbstractMonitorTickets
from cloud_governance.cloud_resource_orchestration.utils.common_operations import get_tag_value_by_name
from cloud_governance.common.clouds.aws.athena.pyathena_operations import PyAthenaOperations
from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.jira.jira_operations import JiraOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import handler, logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class AzureMonitorTickets(AbstractMonitorTickets):
    """This method monitors the Jira Tickets"""

    NEW = 'New'
    REFINEMENT = 'Refinement'
    CLOSED = 'Closed'
    IN_PROGRESS = 'In Progress'
    CLOSE_JIRA_TICKET = 0
    FIRST_CRO_ALERT: int = 5
    SECOND_CRO_ALERT: int = 3
    DEFAULT_ROUND_DIGITS: int = 3

    def __init__(self, region_name: str = ''):
        super().__init__()

    # Todo All the below methods implement in future releases
    def update_budget_tag_to_resources(self, region_name: str, ticket_id: str, updated_budget: int):
        pass

    def update_duration_tag_to_resources(self, region_name: str, ticket_id: str, updated_duration: int):
        pass

    def update_cluster_cost(self):
        pass

    def extend_tickets_budget(self, ticket_id: str, region_name: str, current_budget: int = 0):
        return super().extend_tickets_budget(ticket_id, region_name, current_budget)

    def extend_ticket_duration(self, ticket_id: str, region_name: str, current_duration: int = 0):
        return super().extend_ticket_duration(ticket_id, region_name, current_duration)

    @logger_time_stamp
    def run(self):
        """
        This method run all methods of jira tickets monitoring
        :return:
        # """
        self.monitor_tickets()




from cloud_governance.policy.helpers.azure.azure_policy_operations import AzurePolicyOperations
from cloud_governance.common.utils.utils import Utils


class UnattachedVolume(AzurePolicyOperations):

    RESOURCE_ACTION = "Delete"

    def __init__(self):
        super().__init__()

    def run_policy_operations(self):
        """
        This method returns the list of unattached volumes
        :return:
        :rtype:
        """
        unattached_volumes = []
        available_volumes = self._get_all_volumes()
        active_cluster_ids = self._get_active_cluster_ids()
        for volume in available_volumes:
            tags = volume.get('tags')
            cleanup_result = False
            cluster_tag = self._get_cluster_tag(tags=tags)
            if Utils.equal_ignore_case(volume.get('disk_state'), 'Unattached') and cluster_tag not in active_cluster_ids:
                cleanup_days = self.get_clean_up_days_count(tags=tags)
                cleanup_result = self.verify_and_delete_resource(
                    resource_id=volume.get('id'), tags=tags,
                    clean_up_days=cleanup_days)
                resource_data = self._get_es_schema(resource_id=volume.get('name'),
                                                    user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                    skip_policy=self.get_skip_policy_value(tags=tags),
                                                    cleanup_days=cleanup_days, dry_run=self._dry_run,
                                                    name=volume.get('name'), region=volume.get('location'),
                                                    cleanup_result=str(cleanup_result),
                                                    resource_action=self.RESOURCE_ACTION,
                                                    cloud_name=self._cloud_name,
                                                    resource_type=f"{volume.get('sku', {}).get('tier')} "
                                                                  f"{volume.get('sku', {}).get('name')}",
                                                    resource_state=volume.get('disk_state') if not cleanup_result else "Deleted",
                                                    volume_size=f"{volume.get('disk_size_gb')} GB"
                                                    )
                unattached_volumes.append(resource_data)
            else:
                cleanup_days = 0
            if not cleanup_result:
                self.update_resource_day_count_tag(resource_id=volume.get("id"), cleanup_days=cleanup_days, tags=tags)

        return unattached_volumes



import calendar
import datetime
import time

import pytz
from azure.core.exceptions import HttpResponseError
from azure.mgmt.costmanagement.models import QueryTimePeriod

from cloud_governance.common.clouds.azure.subscriptions.azure_operations import AzureOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


class CostManagementOperations:
    """This class for fetching the azure usage and forecast reports"""

    def __init__(self):
        self.azure_operations = AzureOperations()

    def __get_query_dataset(self, grouping: list, tags: dict, granularity: str):
        """
        This method returns the dataset
        :param grouping:
        :type grouping:
        :param tags:
        :type tags:
        :return:
        :rtype:
        """
        query_dataset = {"aggregation": {"totalCost": {"name": "Cost", "function": "Sum"}},
                         "granularity": granularity,
                         }
        if tags:
            filter_tags = {}
            if len(tags) > 2:
                for key, value in tags.items():
                    and_filter = {'tags': {'name': key.lower(), "operator": "In", 'values': [value.lower()]}}
                    filter_tags.setdefault('and', []).append(and_filter)
            else:
                for key, value in tags.items():
                    filter_tags = {'tags': {'name': key.lower(), "operator": "In", 'values': [value.lower()]}}
            query_dataset['filter'] = filter_tags
        if grouping:
            filter_grouping = []
            for group in grouping:
                if isinstance(group, dict):
                    filter_grouping.append({"name": group['name'], "type": group['type']})
                else:
                    filter_grouping.append({"name": group.lower(), "type": "TagKey"})
            query_dataset['grouping'] = filter_grouping
        return query_dataset

    @logger_time_stamp
    def get_usage(self, scope: str, start_date: datetime = None, end_date: datetime = None,
                  granularity: str = 'Monthly', tags: dict = None, grouping: list = None, **kwargs):
        """
        This method get the current usage based on month
        :param scope:
        :param start_date:
        :param end_date:
        :param granularity:
        :param tags:
        :param grouping:
        :param kwargs:
        :return:
        """
        try:
            if not start_date and not end_date:
                end_date = datetime.datetime.now(pytz.UTC)
                start_date = (end_date - datetime.timedelta(days=30)).replace(day=1)
            response = self.azure_operations.cost_mgmt_client.query.usage(scope=scope, parameters={
                'type': 'Usage', 'timeframe': 'Custom',
                'time_period': QueryTimePeriod(from_property=start_date, to=end_date),
                'dataset': self.__get_query_dataset(grouping=grouping, tags=tags, granularity=granularity)
                })
            return response.as_dict()
        except HttpResponseError as e:
            logger.error(e)
            if e.status_code == 429:
                time.sleep(10)
                return self.get_usage(scope, start_date=start_date, end_date=end_date, granularity=granularity, **kwargs)
        except Exception as err:
            logger.error(err)
        return []

    @logger_time_stamp
    def get_forecast(self, scope: str, start_date: datetime = '', end_date: datetime = '', granularity: str = 'Monthly',
                     tags: dict = None, grouping: list = None, **kwargs):
        """
        This method gets the forecast of next couple of months
        @param start_date:
        @param end_date:
        @param granularity:
        @param scope:
        @param tags:
        @param grouping:
        @return:
        """
        try:
            if not start_date and not end_date:
                current = datetime.datetime.now(pytz.UTC)
                start_date = (current - datetime.timedelta(60)).replace(day=1)
                end_date = (current + datetime.timedelta(365))
                month_end = calendar.monthrange(end_date.year, end_date.month)[1]
                end_date = end_date.replace(day=month_end)
            logger.info(f'StartDate: {start_date}, EndDate: {end_date}')
            response = self.azure_operations.cost_mgmt_client.forecast.usage(scope=scope, parameters={
                        'type': 'ActualCost', 'timeframe': 'Custom',
                        'time_period': QueryTimePeriod(from_property=start_date, to=end_date),
                        'dataset': self.__get_query_dataset(grouping=grouping, tags=tags, granularity=granularity),
                        'include_actual_cost': True, 'include_fresh_partial_cost': False
            }).as_dict()
            result = {'columns': response.get('columns'), 'rows': []}
            row_data = {}
            for data in response.get('rows'):
                data_date = data[1]
                if data_date in row_data:
                    if row_data[data_date][2] == 'Actual' and data[2] == 'Forecast':
                        row_data[data_date][2] = data[2]
                    row_data[data_date][0] += data[0]
                else:
                    row_data[data_date] = data
            result['rows'] = list(row_data.values())
            return result
        except HttpResponseError as e:
            logger.error(e)
            if e.status_code == 429:
                time.sleep(10)
                return self.get_usage(scope, start_date=start_date, end_date=end_date, granularity=granularity, **kwargs)
        except Exception as err:
            logger.error(err)
        return []

    def get_filter_data(self, cost_data: dict, tag_name: str = 'User'):
        """
        This method returns the cost data in dict format
        :param tag_name:
        :type tag_name:
        :param cost_data:
        :type cost_data:
        :return:
        :rtype:
        """
        output_list = self.get_prettify_data(cost_data)
        users_list = {}
        for item in output_list:
            tag_value = item.get('TagValue')
            if tag_value not in users_list:
                users_list[tag_value] = {}
                users_list[tag_value]['Cost'] = item.get('Cost')
            else:
                users_list[tag_value]['Cost'] = users_list[tag_value]['Cost'] + item.get('Cost')
        users_cost = []
        for value, cost in users_list.items():
            users_cost.append({'User': value, 'Cost': cost.get('Cost')})
        return users_cost

    def get_prettify_data(self, cost_data: dict):
        """
        This method returns the prettify data
        :param cost_data:
        :type cost_data:
        :return:
        :rtype:
        """
        columns = cost_data.get('columns')
        columns_data = [column.get('name') for column in columns]
        rows = cost_data.get('rows')
        rows_data = [dict(zip(columns_data, row)) for row in rows]
        return rows_data

    def get_total_cost(self, cost_data: dict):
        """
        This method returns the total cost of the data dict
        :param cost_data:
        :type cost_data:
        :return:
        :rtype:
        """
        output_list = self.get_prettify_data(cost_data)
        total_sum = 0
        for item in output_list:
            total_sum += item.get('Cost')
        return total_sum



from ast import literal_eval

from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload
from cloud_governance.common.clouds.aws.cost_explorer.cost_explorer_operations import CostExplorerOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class CostExplorer:
    """
    This class fetches the cost_explorer report from the AWS based on two days ago data and upload to ElasticSearch.
    fetching AWS cost explorer of two days ago because day ago cost calculation is not closed.
    """

    BULK_UPLOAD_THREADS = 8
    SAVINGS_PLAN_FILTER = {  # savings plan usage for ce filter
                            'Dimensions': {
                                        'Key': 'RECORD_TYPE',
                                        'Values': ['SavingsPlanRecurringFee', 'SavingsPlanNegation', 'SavingsPlanCoveredUsage']
                                }
                    }
    EXCLUDE = 'exclude'
    INCLUDE = 'include'
    CE_KEY_RESULTS_BY_TIME = 'ResultsByTime'
    INDEX_ID = 'index_id'
    APPEND = 'a'
    CE_OPERATION = 'Not'

    def __init__(self):
        super().__init__()
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.start_date = self.__environment_variables_dict.get('start_date', '')  # yyyy-mm-dd
        self.end_date = self.__environment_variables_dict.get('end_date', '')  # yyyy-mm-dd
        self.granularity = self.__environment_variables_dict.get('granularity', 'DAILY')
        self.cost_metric = self.__environment_variables_dict.get('cost_metric', 'UnblendedCost')
        self.cost_tags = literal_eval(self.__environment_variables_dict.get('cost_explorer_tags', '{}'))
        self.file_name = self.__environment_variables_dict.get('file_name', '')
        self.__cost_explorer = CostExplorerOperations()
        self.region = self.__environment_variables_dict.get('AWS_DEFAULT_REGION', 'us-east-1')
        self._ec2_operations = EC2Operations(region=self.region)
        self._elastic_upload = ElasticUpload()
        self.__account = self.__environment_variables_dict.get('account').upper().replace('OPENSHIFT-', "").strip()

    def filter_data_by_tag(self, groups: dict, tag: str, savings_plan: str):
        """
        This method extract data by tag
        @param tag:
        @param groups: Data from the cloud explorer
        @param savings_plan:
        @return: converted into dict format
        """
        data = {}
        start_time = groups.get('TimePeriod').get('Start')
        account = self.__account
        for group in groups.get('Groups'):
            name = ''
            amount = ''
            if group.get('Keys'):
                name = group.get('Keys')[0].split('$')[-1].strip().replace(' ', '-')
                if name == 'PERF-SCALE':
                    name = 'PERFSCALE'
                name = name if name else f'{self._elastic_upload.account}-REFUND/TAX/RECURRING_FEE'
            if group.get('Metrics'):
                amount = group.get('Metrics').get(self.cost_metric).get('Amount')
            if name and amount:
                if 'aws-go-sdk' in name:
                    name = 'aws-go-sdk'
                else:
                    if 'vm_import_image' in name:
                        name = 'vm_import_image'
                index_id = f'{start_time}-{account}-{tag}-savings-{savings_plan}-{name}'.lower()
                if index_id not in data:
                    upload_data = {tag: name if tag.upper() in list(self.__cost_explorer.CE_COST_TYPES) else name.upper(),
                                   'Cost': round(float(amount), 3), self.INDEX_ID: index_id, 'timestamp': start_time, 'savings_plan': savings_plan}
                    if 'global' in self._elastic_upload.es_index:
                        if 'Budget' not in upload_data:
                            upload_data['Budget'] = self._elastic_upload.account
                    upload_data['tag'] = tag.lower()
                    data[index_id] = upload_data
                else:
                    data[index_id]['Cost'] += round(float(amount), 3)
        return list(data.values())

    def __get_daily_cost_by_tags(self):
        """
        This method extracts the costs by tags and upload to elastic search
        @return:
        """
        data_house = {}
        for tag in self.cost_tags:
            for savings_plan in [self.EXCLUDE, self.INCLUDE]:
                filters = {}
                if savings_plan == self.EXCLUDE:
                    filters = {self.CE_OPERATION: self.SAVINGS_PLAN_FILTER}
                if self.start_date and self.end_date:
                    response = self.__cost_explorer.get_cost_by_tags(tag=tag, start_date=self.start_date, end_date=self.end_date, granularity=self.granularity, cost_metric=self.cost_metric, Filter=filters)
                else:
                    response = self.__cost_explorer.get_cost_by_tags(tag=tag, granularity=self.granularity, cost_metric=self.cost_metric, Filter=filters)
                results_by_time = response.get(self.CE_KEY_RESULTS_BY_TIME)
                if results_by_time:
                    data_house[f'{tag}-{savings_plan}'] = []
                    for result in results_by_time:
                        data_house[f'{tag}-{savings_plan}'].extend(self.filter_data_by_tag(result, tag, savings_plan))
        return data_house

    @logger_time_stamp
    def __upload_data(self, data: list, index: str):
        """
        This method upload to elastic search
        @param data:
        @param index:
        @return:
        """
        if self.file_name:
            with open(f'/tmp/{self.file_name}', self.APPEND) as file:
                for value in data:
                    file.write(f'{value}\n')
        else:
            for value in data:
                self.upload_item_to_es(index=index, item=value, index_id=value[self.INDEX_ID])
            logger.info(f'Data uploaded to {index}, Total Data: {len(data)}')

    def upload_item_to_es(self, item: dict, index: str, index_id: str = ''):
        """
        This method upload one item to es
        @param item:
        @param index:
        @param index_id:
        @return:
        """
        if index_id:
            self._elastic_upload.elastic_search_operations.upload_to_elasticsearch(index=index, data=item, id=index_id)
        else:
            self._elastic_upload.elastic_search_operations.upload_to_elasticsearch(index=index, data=item)

    def upload_tags_cost_to_elastic_search(self):
        """
        This method upload daily tag cost into ElasticSearch
        @return:
        """
        logger.info(f'Get {self.granularity} Cost usage by metric: {self.cost_metric}')
        cost_data = self.__get_daily_cost_by_tags()
        for key, values in cost_data.items():
            logger.info(f"Uploading the data of {key} tag")
            self.__upload_data(values, self._elastic_upload.es_index)

    def run(self):
        """
        This method run the operations
        @return:
        """
        self.upload_tags_cost_to_elastic_search()
